// SPDX-License-Identifier: GPL-3.0-or-later
//
// bdc_hashtable.c
//
// Copyright (C) 2011-2024, University of Vienna and Vienna Institute for Nature Conservation & Analyses, Andreas Gattringer.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
//

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "bdc_hashtable.h"
#include "bdc_logging/logging.h"
#include "bdc_memory/bdc_memory.h"


struct bdc_value {
        union {
                char *string;
                int64_t int64;
                double double_value;
                void *pointer;
        };
        enum bdc_value_type type;
};
struct bdc_item {
        char *key;
        struct bdc_value *value;
        struct bdc_item *next;
};
struct bdc_hash_table {
        size_t item_count;
        size_t bucket_count;
        size_t *items_per_bucket;
        //size_t max_items_pro_bucket;
        struct bdc_item **items;

};



// FNV-1a hash, see https://en.wikipedia.org/wiki/Fowler%E2%80%93Noll%E2%80%93Vo_hash_function#FNV-1a_hash
#define OFFSET 14695981039346656037UL
#define PRIME 1099511628211UL

#define DEBUG(...)

// FIXME: add function to return keys

uint64_t hash_function(const char *key)
{
        uint64_t hash = OFFSET;
        while (*key != '\0') {
                hash ^= (uint64_t) (unsigned char) *key;
                hash *= PRIME;
                key++;
        }
        return hash;
}


__attribute__((unused)) size_t hash_index(const char *key, size_t buckets)
{
        return (size_t) hash_function(key) & (buckets - 1);
}


size_t hash_idx_from_hash(uint64_t hash, size_t buckets)
{
        return hash & (buckets - 1);
}



struct bdc_hash_table *bdc_intern_new_hash_table(size_t buckets)
{
        struct bdc_hash_table *table = calloc_or_die(1, sizeof(struct bdc_hash_table));
        table->bucket_count = buckets;
        table->items = calloc_or_die(table->bucket_count, sizeof(struct bdc_item *));
        table->item_count = 0;
        //table->max_items_pro_bucket = 0;

        table->items_per_bucket = calloc_or_die(table->bucket_count, sizeof(size_t));
        DEBUG("Creating hash table with %lu buckets\n", table->bucket_count)

        for (size_t i = 0; i < table->bucket_count; i++) {
                table->items_per_bucket[i] = 0;
                table->items[i] = NULL;
        }

        return table;
}


__attribute__((unused)) struct bdc_hash_table *new_hash_table(void)
{
        return bdc_intern_new_hash_table(MIN_SIZE);
}


void destroy_value(struct bdc_value **value)
{
        if (value == NULL || *value == NULL) return;
        struct bdc_value *this = *value;
        if(this->type == BDC_STRING) {
                log_message(LOG_DEBUG, "freeing string value '%s'", this->string);
                free(this->string);
                this->string = NULL;
        }
        log_message(LOG_DEBUG, "freeing value");
        free(this);
        *value = NULL;

}

struct bdc_item *bdc_intern_get_entry(struct bdc_hash_table *table, const char *key, uint64_t hash)
{
        size_t hash_idx = hash_idx_from_hash(hash, table->bucket_count);
        if (table->items[hash_idx] == NULL) return NULL;

        struct bdc_item *entry = table->items[hash_idx];

        if (0 == strcmp(entry->key, key)) return entry;

        while (entry->next != NULL) {
                entry = entry->next;
                if (0 == strcmp(entry->key, key)) return entry;
        }

        return NULL;
}


struct bdc_item *get_entry(struct bdc_hash_table *table, const char *key)
{
        uint64_t hash = hash_function(key);

        size_t hash_idx = hash & (table->bucket_count - 1);
        if (table->items[hash_idx] == NULL) return NULL;

        struct bdc_item *entry = table->items[hash_idx];
        if (0 == strcmp(entry->key, key)) return entry;
        while (entry->next != NULL) {
                entry = entry->next;
                if (0 == strcmp(entry->key, key)) return entry;
        }

        return NULL;
}

void destroy_item(struct bdc_item **item)
{
        if(item == NULL || *item == NULL) return;
        struct bdc_item *this = *item;
        destroy_value(&this->value);
        free(this->key);
        free(this);
        *item = NULL;
}

bool remove_key(struct bdc_hash_table *table, const char *key)
{
        if (!bdc_key_in_hash_table(table, key)) return false;

        uint64_t hash = hash_function(key);
        size_t hash_idx = hash & (table->bucket_count - 1);

        struct bdc_item *target = NULL;
        struct bdc_item *entry = table->items[hash_idx];
        if (0 == strcmp(entry->key, key)) {
                target = entry;
                table->items[hash_idx] = target->next;
                table->items_per_bucket[hash_idx]--;
                destroy_item(&target);
                return true;
        }
        struct bdc_item *prev = NULL;
        while (entry->next != NULL) {
                prev = entry,
                        entry = entry->next;
                if (0 == strcmp(entry->key, key)) {
                        target = entry;
                        prev->next = target->next;
                        table->items_per_bucket[hash_idx]--;
                        free(entry);
                        return true;
                }
        }
        return false;
}


__attribute__((unused)) bool bdc_key_in_hash_table(struct bdc_hash_table *table, const char *key)
{
        return get_entry(table, key) != NULL;

}


struct bdc_value *bdc_get_value_from_key(struct bdc_hash_table *table, const char *key)
{
        struct bdc_item *entry = get_entry(table, key);
        if (entry) {
                return entry->value;
        } else {
                return NULL;
        }
}


struct bdc_item *new_entry(const char *key, struct bdc_value *value)
{
        struct bdc_item *entry = calloc_or_die(1, sizeof(struct bdc_item));


        entry->value = value;
        log_message(LOG_DEBUG, "duplicating key '%s'", key);
        entry->key = strdup(key);
        entry->next = NULL;

        return entry;
}


void print_key_val(struct bdc_item *entry)
{
        struct bdc_value *value = entry->value;
        switch (value->type) {

                case BDC_STRING:
                        printf("\t\t%s -> '%s'\n", entry->key, value->string);
                        break;
                case BDC_INT64:
                        printf("\t\t%s -> %ld\n", entry->key, value->int64);
                        break;
                case BDC_DOUBLE:
                        printf("\t\t%s -> %f\n", entry->key, value->double_value);
                        break;
                case BDC_CUSTOM:
                        printf("\t\t%s -> %p\n", entry->key, value->pointer);
                        break;
        }

}


__attribute__((unused)) void print_hash_table(const struct bdc_hash_table *table)
{
        printf("\n");
        printf("Hash table: %lu buckets with %lu items\n", table->bucket_count,
               table->item_count);
        for (size_t i = 0; i < table->bucket_count; i++) {
                if (table->items[i] == NULL) continue;
                printf("\tentry %lu: %lu entries\n", i, table->items_per_bucket[i]);
                size_t count = 1;
                struct bdc_item *entry = table->items[i];
                print_key_val(entry);
                while (entry->next != NULL) {
                        entry = entry->next;
                        count++;
                        print_key_val(entry);
                }

                if (count != table->items_per_bucket[i]) {
                        printf("Error: mismatch between expected (%lu) and found entries (%lu)\n",
                               table->items_per_bucket[i], i);
                }
        }
        printf("\n");
}





struct bdc_item *bdc_intern_add_entry(struct bdc_hash_table *table, struct bdc_item *entry, uint64_t hash)
{
        size_t hash_idx = hash_idx_from_hash(hash, table->bucket_count);
        entry->next = table->items[hash_idx];
        table->items[hash_idx] = entry;
        table->items_per_bucket[hash_idx] += 1;
        table->item_count += 1;

        return entry;
}


void rehash_table(struct bdc_hash_table *table)
{
        log_message(LOG_INFO, "%s: rehashing!", __func__);
        struct bdc_hash_table *tmp = bdc_intern_new_hash_table(table->bucket_count * 2);

        for (size_t i = 0; i < table->bucket_count; i++) {
                if (table->items[i] == NULL) continue;
                struct bdc_item *entry = table->items[i];
                struct bdc_item *next = entry->next;
                uint64_t hash = hash_function(entry->key);
                bdc_intern_add_entry(tmp, entry, hash);

                while (next) {
                        entry = next;
                        next = entry->next;
                        hash = hash_function(entry->key);
                        bdc_intern_add_entry(tmp, entry, hash);
                }

        }

        free(table->items);
        free(table->items_per_bucket);
        table->items = tmp->items;
        table->items_per_bucket = tmp->items_per_bucket;
        table->bucket_count = tmp->bucket_count;

        free(tmp);
        tmp = NULL;

}


struct bdc_item *bdc_ht_add_string(struct bdc_hash_table *table, const char *key, const char *value)
{
        struct bdc_value *val = calloc(1, sizeof(struct bdc_value));
        log_message(LOG_DEBUG, "duplicating value '%s", value);
        val->string = strdup(value);
        val->type = BDC_STRING;
        return intern_bdc_add_entry(table, key, val);
}


struct bdc_item *bdc_ht_add_custom(struct bdc_hash_table *table, const char *key, void *value)
{
        struct bdc_value *val = calloc(1, sizeof(struct bdc_value));
        val->pointer = value;
        val->type = BDC_CUSTOM;
        return intern_bdc_add_entry(table, key, val);
}


struct bdc_item *bdc_ht_add_int64(struct bdc_hash_table *table, const char *key, int64_t value)
{
        struct bdc_value *val = calloc(1, sizeof(struct bdc_value));
        val->int64 = value;
        val->type = BDC_INT64;
        return intern_bdc_add_entry(table, key, val);
}


struct bdc_item *bdc_ht_add_double(struct bdc_hash_table *table, const char *key, double value)
{
        struct bdc_value *val = calloc(1, sizeof(struct bdc_value));
        val->double_value = value;
        val->type = BDC_DOUBLE;
        return intern_bdc_add_entry(table, key, val);
}


struct bdc_item *intern_bdc_add_entry(struct bdc_hash_table *table, const char *key, struct bdc_value *value)
{
        uint64_t hash = hash_function(key);
        size_t hash_idx = hash_idx_from_hash(hash, table->bucket_count);
        switch (value->type) {
                case BDC_STRING:
                        log_message(LOG_DEBUG, "adding entry %s %s with hash %lu to bucket %lu", key, value->string,
                                    hash_function(key), hash_idx);
                        break;
                case BDC_INT64:
                        log_message(LOG_DEBUG, "adding entry %s %ld with hash %lu to bucket %lu", key, value->int64,
                                    hash_function(key), hash_idx);
                        break;
                case BDC_DOUBLE:
                        log_message(LOG_DEBUG, "adding entry %s %f with hash %lu to bucket %lu", key,
                                    value->double_value,
                                    hash_function(key), hash_idx);
                        break;
                case BDC_CUSTOM:
                        log_message(LOG_DEBUG, "adding entry %s %p with hash %lu to bucket %lu", key, value->pointer,
                                    hash_function(key), hash_idx);
                        break;
        }


        struct bdc_item *entry = bdc_intern_get_entry(table, key, hash);

        if (entry != NULL) {
                log_message(LOG_DEBUG, "replacing key '%s'", entry->key);
                destroy_value(&entry->value);
                entry->value = value;

                if (entry->value->type == BDC_STRING) {
                        log_message(LOG_DEBUG, "setting value of '%s' to '%s'", key, entry->value->string);
                }
                return entry;
        }
        assert(entry == NULL);
        entry = new_entry(key, value);
        bdc_intern_add_entry(table, entry, hash);

        if (table->items_per_bucket[hash_idx] > 2) {
                rehash_table(table);
        }

        log_message(LOG_DEBUG, "added key '%s'\n", key);
        return entry;
}


bool bdc_ht_get_int64_from_key(struct bdc_hash_table *table, const char *key, int64_t *result)
{
        struct bdc_value *value = bdc_get_value_from_key(table, key);
        if (!value) return false;
        if (value->type != BDC_INT64) return false;
        *result = value->int64;
        return true;
}


__attribute__((unused)) bool bdc_ht_get_double_from_key(struct bdc_hash_table *table, const char *key, double *result)
{
        struct bdc_value *value = bdc_get_value_from_key(table, key);
        if (!value) return false;
        if (value->type != BDC_DOUBLE) return false;
        *result = value->double_value;
        return true;
}


bool bdc_ht_get_string_from_key(struct bdc_hash_table *table, const char *key, char **result)
{
        struct bdc_value *value = bdc_get_value_from_key(table, key);
        if (!value) return false;
        if (value->type != BDC_STRING) return false;
        *result = strdup(value->string);
        return true;
}

void destroy_hash_table(struct bdc_hash_table **ht)
{
        if (ht == NULL || *ht == NULL) return;
        struct bdc_hash_table *t = *ht;
        size_t count = t->bucket_count;
        for (size_t i = 0; i < count; i++){
                struct bdc_item *item = t->items[i];
                while(item){
                        struct bdc_item *this = item;
                        item = item->next;
                        if (this->value->type == BDC_STRING) {
                                log_message(LOG_DEBUG, "freeing item '%s' -> '%s'", this->key, this->value->string);
                                free(this->value->string);
                        }
                        free(this->value);
                        free(this->key);
                        free(this);
                        this = NULL;
                }
                free(item);
        }
        free(t->items_per_bucket);
        free(t->items);
        free(t);
        *ht = NULL;
}