// SPDX-License-Identifier: GPL-3.0-or-later
//
// bdc_paths.c
//
// Copyright (C) 2011-2024, University of Vienna and Vienna Institute for Nature Conservation & Analyses, Andreas Gattringer.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
//

#include <string.h>
#include <sys/stat.h>
#include <stdio.h>
#include <errno.h>

#include "bdc_io.h"
#include "bdc_cross_platform/bdc_cross_platform.h"

#ifdef BDC_ON_WINDOWS
#include <direct.h>
#else
#include <unistd.h>
#endif



bool directory_exists(const char *path)
{
        struct stat info;
        int rc = stat(path, &info);
        if (rc == 0 && info.st_mode & S_IFDIR) return true;

        if (rc == 0) {
                fprintf(stderr, "Path '%s' is not a directory. Exiting.\n", path);
                exit(EXIT_FAILURE);
        }

        fprintf(stderr, "Directory '%s' does not exist\n", path);
        return false;
}


bool directory_writable(const char *path)
{
        int rc = access(path, F_OK);

        if (rc == 0) {
                return true;
        }

        fprintf(stderr, "directory %s exists but cannot be accessed or written to.\n", path);
        return false;
}


void make_directory(const char *path)
{
#ifndef BDC_ON_WINDOWS
        int rc = mkdir(path, S_IRWXU);
#else
        int rc = _mkdir(path);
#endif
        if (rc == 0) return;

        int error = errno;
        fprintf(stderr, "unable to create directory %s: %s\n", path, strerror(error));
        exit(EXIT_FAILURE);
}


char *assemble_path(const struct string_array *path)
{
        return string_array_paste(path, DIRECTORY_SEPARATOR);
}


char *
assemble_filename(struct string_array *path, struct string_array *filename, char *field_separator, char *extension)
{
        char *name_string = string_array_paste(filename, field_separator);
        char *result = NULL;

        if (path && get_string_count(path) > 0) {
                char *path_string = string_array_paste(path, DIRECTORY_SEPARATOR);
                if (extension && strlen(extension)) {
                        char *fn = compound_string(name_string, extension, ".");
                        result = compound_string(path_string, fn, DIRECTORY_SEPARATOR);
                        free(fn);
                } else {
                        result = compound_string(path_string, name_string, DIRECTORY_SEPARATOR);
                }
                free(path_string);
        } else {

                if (extension && strlen(extension)) {
                        result = compound_string(name_string, extension, ".");
                } else {
                        result = strdup(name_string);
                }
        }

        free(name_string);
        return result;
}


bool check_and_create_directory_if_needed(const struct string_array *directory_path)
{
        if (directory_path == NULL || get_string_count(directory_path) == 0) {
                fprintf(stderr, "Unable to check directory - no path specified\n");
                exit(EXIT_FAILURE);
        }

        char *dir = assemble_path(directory_path);


        if (!directory_exists(dir)) {
                fprintf(stderr, "directory '%s' does not exist. Trying to create.\n", dir);
                make_directory(dir);
        }

        bool writeable = directory_writable(dir);
        free(dir);
        return writeable;
}