// SPDX-License-Identifier: GPL-3.0-or-later
//
// bdc_strings.h
//
// Copyright (C) 2011-2024, University of Vienna and Vienna Institute for Nature Conservation & Analyses, Andreas Gattringer.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
//

#ifndef CATS_STRING_H_
#define CATS_STRING_H_

#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>

struct string_array;
struct string_substitution;
enum JSON_HINT {
        JSON_STRING = 0,
        JSON_QUOTED = 0,
        JSON_UNQUOTED = 1,
        JSON_NUMERIC = 1
};
// constants
#define TEXT_DIVIDER "=================================================================================================="
#define PAD_NAME 55
#define PAD_VALUE 35

//#define DEBUG_STRING
#ifdef DEBUG_STRING
#define DBG_STRING(X) {X;}while(0);
#else
#define DBG_STRING(X)
#endif


// string array functions
struct string_array *new_string_array(void);

const char *get_string(const struct string_array *array, int count);

int get_string_count(const struct string_array *array);

void free_string_array(struct string_array **array);

struct string_array *get_all_tokens(const char *line, const char *sep);


void string_array_add(struct string_array *array, const char *string);

void string_array_add_int64(struct string_array *array, int64_t number, char *format_string);

char *string_array_paste(const struct string_array *array, const char *sep);


void string_array_add_int(struct string_array *array, int32_t number, const char *format_string);

struct string_array *new_string_array_init(const char *entry);


__attribute__((unused)) void string_array_add_conditional(struct string_array *array, char *string, bool conditon);

__attribute__((unused)) void string_array_add_double(struct string_array *array, double number, const char *format_string);

__attribute__((unused)) void
string_array_add_int_conditional(struct string_array *array, int32_t number, const char *format_string, bool condition);

__attribute__((unused)) const char *string_array_get_string(const struct string_array *array, int n);

__attribute__((unused)) void print_string_array(const struct string_array *array);


struct string_array *copy_string_array(const struct string_array *src);

char *json_dict(const struct string_array *header, const struct string_array *data);

int32_t string_array_string_index(const struct string_array *array, const char *needle);

__attribute__((unused)) bool string_array_contains_string(const struct string_array *array, const char *needle);


// string functions
char *trim_both(char *line);

int32_t count_fields(const char *line, const char *sep);


char *remove_0th_token(const char *line, const char *sep);

__attribute__((unused)) char *get_nth_token(char *line, const char *sep, int32_t n);

size_t substring_count(const char *haystack, const char *needle);

char *compound_string(const char *s1, const char *s2, const char *sep);

char *replace_substring(const char *original, const char *search, const char *replacement);

bool string_is_emptyish(const char *string);

char *empty_string(void);

char *right_trim(char *line);

char *left_trim(char *line);

bool string_is_empty_or_null(const char *name);

void abort_on_null_token(const char *token, const char *orig, int count);
// helper functions

void asprintf_check(int rc);

void snprintf_check(int rc, size_t expected);


// pretty print functions
void print_string(int indent, const char *name, const char *value);

void print_subcategory(int indent, const char *category_name, int64_t category_number, const char *name);

void print_category(int indent, const char *category, const char *name);

void print_integer(int indent, const char *name, int64_t value);

void print_rate(int indent, const char *name, long double value);



// string substitution

void add_substitution(struct string_substitution *substituter, const char *key, const char *initial_value);

void update_substitution(struct string_substitution *substituter, const char *key, const char *initial_value);

struct string_substitution *new_substituter(void);

char *substitute_string(const char *string, const struct string_substitution *substituter);

void destroy_substituter(struct string_substitution **substituter);

struct string_substitution *copy_substituter(const struct string_substitution *from);

void add_substitution_integer(struct string_substitution *substituter, const char *key, int32_t initial_value);

void set_integer_pattern(struct string_substitution *substituter, const char *key, const char *pattern);


// conversion function
const char *bool_to_string(bool b);

bool string_to_float(char *string, float *value);

bool string_to_bool(char *string, bool *value);

bool string_to_double(const char *string, double *value);

bool string_to_integer(const char *string, int32_t *value);

bool string_to_long_double(const char *string, long double *value);

#endif
