// SPDX-License-Identifier: GPL-3.0-or-later
//
// string_substitution.c
//
// Copyright (C) 2011-2024, University of Vienna and Vienna Institute for Nature Conservation & Analyses, Andreas Gattringer.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
//

#include <string.h>

#include "bdc_memory/bdc_memory.h"
#include "bdc_logging/logging.h"
#include "bdc_hashtable/bdc_hashtable.h"
#include "bdc_strings.h"

struct string_substitution {
        struct bdc_hash_table *substitution_strings;
        struct bdc_hash_table *integer_patterns;
        struct string_array *keys;
};

struct string_substitution *new_substituter()
{
        struct string_substitution *new = calloc_or_die(1, sizeof(struct string_substitution));
        new->substitution_strings = new_hash_table();
        new->integer_patterns = new_hash_table();
        new->keys = new_string_array();
        return new;
}


void set_integer_pattern(struct string_substitution *substituter, const char *key, const char *pattern)
{
        if (substituter == NULL) {
                log_message(LOG_ERROR, "%s: substituter == NULL", __func__);
                exit(EXIT_FAILURE);
        }
        bdc_ht_add_string(substituter->integer_patterns, key, pattern);
}

void add_substitution_integer(struct string_substitution *substituter, const char *key, int32_t initial_value)
{
        if (substituter == NULL) {
                log_message(LOG_ERROR, "%s: substituter == NULL", __func__);
                exit(EXIT_FAILURE);
        }

        char *default_pattern = "%d";
        char *pattern = NULL;
        char *loaded_pattern = NULL;
        bdc_ht_get_string_from_key(substituter->integer_patterns, key, &loaded_pattern);
        if (!loaded_pattern) {
                pattern = default_pattern;

        } else {
                pattern = loaded_pattern;
        }

        char *value = NULL;
        int rc = asprintf(&value, pattern, initial_value);
        asprintf_check(rc);
        add_substitution(substituter, key, value);
        free(value);
        free(loaded_pattern);
}





void add_substitution(struct string_substitution *substituter, const char *key, const char *initial_value)
{
        if (!bdc_key_in_hash_table(substituter->substitution_strings, key)) {
                string_array_add(substituter->keys, key);
        }

        bdc_ht_add_string(substituter->substitution_strings, key, initial_value);
}

char *substitute_string(const char *string, const struct string_substitution *substituter)
{
        if (substituter == NULL) {
                log_message(LOG_ERROR, "%s: substituter == NULL", __func__);
                exit(EXIT_FAILURE);
        }
        if (substituter->keys == NULL) {
                log_message(LOG_ERROR, "%s: substituter invalid, keys = NULL", __func__);
                exit(EXIT_FAILURE);
        }
        log_message(LOG_INFO, "substituting string '%s' with %d potential substitutions", string, get_string_count(substituter->keys));
        const int count = get_string_count(substituter->keys);
        char *copy = strdup(string);
        char *result = strdup(string);
        for (int i = 0; i < count; i++) {
                const char *key = get_string(substituter->keys, i);
                char *value = NULL;
                bool found = bdc_ht_get_string_from_key(substituter->substitution_strings, key, &value);
                if (! found) {
                        continue;
                }
                free(result);
                result = replace_substring(copy, key, value);
                free(copy);
                copy = strdup(result);


        }
        free(copy);
        return result;
}

void destroy_substituter(struct string_substitution **substituter)
{
        if (substituter == NULL || *substituter == NULL) return;
        struct string_substitution *s = *substituter;
        free_string_array(&s->keys);
        destroy_hash_table(&s->substitution_strings);
        destroy_hash_table(&s->integer_patterns);
        free(s);
        *substituter = NULL;

}

struct string_substitution *copy_substituter(const struct string_substitution *from)
{
        if (from == NULL) return NULL;
        struct string_substitution *to = new_substituter();
        const struct string_array *keys = from->keys;
        const int32_t count = get_string_count(keys);
        for (int32_t i = 0; i< count; i++) {
                const char *key = get_string(keys, i);
                char *value = NULL;
                bdc_ht_get_string_from_key(from->substitution_strings, key, &value);
                if (value) {
                        add_substitution(to, key, value);
                }
                bdc_ht_get_string_from_key(from->integer_patterns, key, &value);
                if (value) {
                        set_integer_pattern(to, key, value);
                }

        }

        return to;
}